/* adept.cpp -- Fast automatic differentiation with expression templates

    Copyright (C) 2012 Robin Hogan <r.j.hogan@reading.ac.uk>

    This file is part of the Adept library.

    This library is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <iostream>
#include <cstring> // For memcpy

#include "adept.h"

namespace adept {

  // Global pointers to the current thread, the second of which is
  // thread safe. The first is only used if ADEPT_STACK_THREAD_UNSAFE
  // is defined.
  __thread Stack* _stack_current_thread = 0;
  Stack* _stack_current_thread_unsafe = 0;


  //using namespace adept;

  // Get a string representation of the error code
  std::string error_message(Status status) {
    switch (status) {
      case SUCCESS:
	return "No error";
	break;
      case FAILURE_GRADIENT_OUT_OF_RANGE:
	return "Attempt to access gradient(s) out of range";
	break;
      case FAILURE_GRADIENT_NOT_INITIALIZED:
	return "Attempt to access uninitialized gradient(s)";
	break;
      case FAILURE_NO_INITIAL_GRADIENTS:
	return "Attempt to perform forward or reverse pass without first setting some gradients";
	break;
      case FAILURE_STACK_ALREADY_ACTIVE:
	return "Attempt to start stack when one is already active in this thread";
	break;
      case FAILURE_XXDEPENDENTS_NOT_IDENTIFIED:
	return "Attempt to compute Jacobian when independents and/or dependents not identified";
	break;
      default:
	break;
    }
    return "Error code not defined";
  }
  
  // Destructor: frees dynamically allocated memory (if any)
  Stack::~Stack() {
    // If this is the currently active stack then set to NULL as
    // "this" is shortly to become invalid
    if (is_thread_unsafe_) {
      if (_stack_current_thread_unsafe == this) {
	_stack_current_thread_unsafe = 0; 
      }
    }
    else if (_stack_current_thread == this) {
      _stack_current_thread = 0; 
    }
#ifdef ADEPT_STACK_STORAGE_DYNAMIC
    if (statement_) {
      delete[] statement_;
    }
    if (gradient_) {
      delete[] gradient_;
    }
    if (multiplier_) {
      delete[] multiplier_;
    }
    if (offset_) {
      delete[] offset_;
    }
#endif
  }
  

  Status
  Stack::start(Offset n)
  {
    // Check that we don't already have an active stack in this thread
    if ((is_thread_unsafe_ && _stack_current_thread_unsafe 
	 && _stack_current_thread_unsafe != this)
	|| ((!is_thread_unsafe_) && _stack_current_thread
	    && _stack_current_thread != this)) {
      return FAILURE_STACK_ALREADY_ACTIVE;
    }
    else {
      if (!is_thread_unsafe_) {
	_stack_current_thread = this;
      }
      else {
	_stack_current_thread_unsafe = this;
      }
      
#ifdef ADEPT_STACK_STORAGE_STL
      statement_.reserve(n);
      multiplier_.reserve(n);
      offset_.reserve(n);
#elif ADEPT_STACK_STORAGE_DYNAMIC
      if (n > n_allocated_operations_) {
	if (multiplier_) {
	  delete[] multiplier_;
	}
	if (offset_) {
	  delete[] offset_;
	}
	multiplier_ = new Real[n];
	offset_ = new Offset[n];
	n_allocated_operations_ = n;
      }
      if (n > n_allocated_statements_) {
	if (statement_) {
	  delete[] statement_;
	}
	statement_ = new Statement[n];
	n_allocated_statements_ = n;
      }
#endif
      clear();
      // Insert a null statement
      push_lhs(-1);
      //    statement_[0].offset = -1;
      //    statement_[0].end_plus_one = 0;
      return SUCCESS;
    }
  }
  
#ifdef ADEPT_STACK_STORAGE_DYNAMIC
  // Double the size of the operation stack, or grow it even more if
  // the requested minimum number of extra entries (min) is greater
  // than this would allow
  void
  Stack::grow_operation_stack(Offset min)
  {
    Offset new_size = 2*n_allocated_operations_;
    if (min > 0 && new_size < n_allocated_operations_+min) {
      new_size += min;
    }
    Real* new_multiplier = new Real[new_size];
    Offset* new_offset = new Offset[new_size];
    
    std::memcpy(new_multiplier, multiplier_, n_operations_*sizeof(Real));
    std::memcpy(new_offset, offset_, n_operations_*sizeof(Offset));
    
    delete[] multiplier_;
    delete[] offset_;
    
    multiplier_ = new_multiplier;
    offset_ = new_offset;
    
    n_allocated_operations_ = new_size;
  }

  // ... likewise for the statement stack
  void
  Stack::grow_statement_stack(Offset min)
  {
    Offset new_size = 2*n_allocated_statements_;
    if (min > 0 && new_size < n_allocated_statements_+min) {
      new_size += min;
    }
    Statement* new_statement = new Statement[new_size];
    std::memcpy(new_statement, statement_,
		n_statements_*sizeof(Statement));
    delete[] statement_;

    statement_ = new_statement;

    n_allocated_statements_ = new_size;
  }
#endif


  // Perform to adjoint computation (reverse mode). It is assumed that
  // some gradients have been assigned already, otherwise the function
  // returns with an error.
  Status
  Stack::compute_adjoint()
  {
    Status status = SUCCESS;
    if (!gradients_are_initialized()) {
      status = FAILURE_NO_INITIAL_GRADIENTS;
    }
    // Loop backwards through the derivative statements
    for (Offset ist = n_statements_-1; ist > 0; ist--) {
      const Statement& statement = statement_[ist];
      // We copy the RHS gradient (LHS in the original derivative
      // statement but swapped in the adjoint equivalent) to "a" in
      // case it appears on the LHS in any of the following statements
      Real a = gradient_[statement.offset];
      gradient_[statement.offset] = 0.0;
      // By only looping if a is non-zero we gain a significant speed-up
      if (a != 0.0) {
	// Loop over operations
	for (Offset i = statement_[ist-1].end_plus_one;
	     i < statement.end_plus_one; i++) {
	  gradient_[offset_[i]] += multiplier_[i]*a;
	}
      }
    }
    return status;
  }
  
  // Perform tangent linear computation (forward mode). It is assumed
  // that some gradients have been assigned already, otherwise the
  // function returns with an error.
  Status
  Stack::compute_tangent_linear()
  {
    Status status = SUCCESS;
    if (!gradients_are_initialized()) {
      status = FAILURE_NO_INITIAL_GRADIENTS;
    }
    else {
      // Loop forward through the statements
      for (Offset ist = 1; ist < n_statements_; ist++) {
	const Statement& statement = statement_[ist];
	// We copy the LHS to "a" in case it appears on the RHS in any
	// of the following statements
	Real a = 0.0;
	for (Offset i = statement_[ist-1].end_plus_one;
	     i < statement.end_plus_one; i++) {
	  a += multiplier_[i]*gradient_[offset_[i]];
	}
	gradient_[statement.offset] = a;
      }
    }
    return status;
  }

  // Compute the Jacobian matrix; note that jacobian_out must be
  // allocated to be of size m*n, where m is the number of dependent
  // variables and n is the number of independents. The independents
  // and dependents must have already been identified with the
  // functions "independent" and "dependent", otherwise this function
  // will fail with FAILURE_XXDEPENDENT_NOT_IDENTIFIED. In the
  // resulting matrix, the "m" dimension of the matrix varies
  // fastest. This is implemented using a forward pass, appropriate
  // for m>n.
  Status
  Stack::jacobian_forward(Real* jacobian_out)
  {
    if (independent_offset_.empty() || dependent_offset_.empty()) {
      return FAILURE_XXDEPENDENTS_NOT_IDENTIFIED;
    }
    gradient_multipass_.resize(max_gradient_);
    // For optimization reasons, we process a block of
    // ADEPT_MULTIPASS_SIZE columns of the Jacobian at once; calculate
    // how many blocks are needed and how many extras will remain
    Offset n_block = n_independent() / ADEPT_MULTIPASS_SIZE;
    Offset n_extra = n_independent() % ADEPT_MULTIPASS_SIZE;
    Offset i_independent = 0; // Index of first column in the block we
			      // are currently computing
    // Loop over blocks of ADEPT_MULTIPASS_SIZE columns
    for (Offset iblock = 0; iblock < n_block; iblock++) {
      // Set the initial gradients all to zero
      zero_gradient_multipass();
      // Each seed vector has one non-zero entry of 1.0
      for (Offset i = 0; i < ADEPT_MULTIPASS_SIZE; i++) {
	gradient_multipass_[independent_offset_[i_independent+i]][i] = 1.0;
      }
      // Loop forward through the derivative statements
      for (Offset ist = 1; ist < n_statements_; ist++) {
	const Statement& statement = statement_[ist];
	// We copy the LHS to "a" in case it appears on the RHS in any
	// of the following statements
	Block<ADEPT_MULTIPASS_SIZE,Real> a; // Initialized to zero
					    // automatically
	// Loop through operations
	for (Offset iop = statement_[ist-1].end_plus_one;
	     iop < statement.end_plus_one; iop++) {
	  // Loop through columns within this block; we hope the
	  // compiler can optimize this loop
	  for (Offset i = 0; i < ADEPT_MULTIPASS_SIZE; i++) {
	    a[i] += multiplier_[iop]*gradient_multipass_[offset_[iop]][i];
	  }
	}
	// Copy the results
	for (Offset i = 0; i < ADEPT_MULTIPASS_SIZE; i++) {
	  gradient_multipass_[statement.offset][i] = a[i];
	}
      } // End of loop over statements
      // Copy the gradients corresponding to the dependent variables
      // into the Jacobian matrix
      for (Offset idep = 0; idep < n_dependent(); idep++) {
	for (Offset i = 0; i < ADEPT_MULTIPASS_SIZE; i++) {
	  jacobian_out[(i_independent+i)*n_dependent()+idep] 
	    = gradient_multipass_[dependent_offset_[idep]][i];
	}
      }
      i_independent += ADEPT_MULTIPASS_SIZE;
    } // End of loop over blocks
    
    // Now do the same but for the remaining few columns in the matrix
    if (n_extra > 0) {
      zero_gradient_multipass();
      for (Offset i = 0; i < n_extra; i++) {
	gradient_multipass_[independent_offset_[i_independent+i]][i] = 1.0;
      }
      for (Offset ist = 1; ist < n_statements_; ist++) {
	const Statement& statement = statement_[ist];
	Block<ADEPT_MULTIPASS_SIZE,Real> a;
	for (Offset iop = statement_[ist-1].end_plus_one;
	     iop < statement.end_plus_one; iop++) {
	  for (Offset i = 0; i < n_extra; i++) {
	    a[i] += multiplier_[iop]*gradient_multipass_[offset_[iop]][i];
	  }
	}
	for (Offset i = 0; i < n_extra; i++) {
	  gradient_multipass_[statement.offset][i] = a[i];
	}
      }
      for (Offset idep = 0; idep < n_dependent(); idep++) {
	for (Offset i = 0; i < n_extra; i++) {
	  jacobian_out[(i_independent+i)*n_dependent()+idep] 
	    = gradient_multipass_[dependent_offset_[idep]][i];
	}
      }
    }
    
    return SUCCESS;
  }


  // Compute the Jacobian matrix; note that jacobian_out must be
  // allocated to be of size m*n, where m is the number of dependent
  // variables and n is the number of independents. The independents
  // and dependents must have already been identified with the
  // functions "independent" and "dependent", otherwise this function
  // will fail with FAILURE_XXDEPENDENT_NOT_IDENTIFIED. In the
  // resulting matrix, the "m" dimension of the matrix varies
  // fastest. This is implemented using a reverse pass, appropriate
  // for m<n.
  Status
  Stack::jacobian_reverse(Real* jacobian_out)
  {
    if (independent_offset_.empty() || dependent_offset_.empty()) {
      return FAILURE_XXDEPENDENTS_NOT_IDENTIFIED;
    }
    gradient_multipass_.resize(max_gradient_);
    // For optimization reasons, we process a block of
    // ADEPT_MULTIPASS_SIZE rows of the Jacobian at once; calculate
    // how many blocks are needed and how many extras will remain
    Offset n_block = n_dependent() / ADEPT_MULTIPASS_SIZE;
    Offset n_extra = n_dependent() % ADEPT_MULTIPASS_SIZE;
    Offset i_dependent = 0; // Index of first row in the block we are
			    // currently computing
    // Loop over the of ADEPT_MULTIPASS_SIZE rows
    for (Offset iblock = 0; iblock < n_block; iblock++) {
      // Set the initial gradients all to zero
      zero_gradient_multipass();
      // Each seed vector has one non-zero entry of 1.0
      for (Offset i = 0; i < ADEPT_MULTIPASS_SIZE; i++) {
	gradient_multipass_[dependent_offset_[i_dependent+i]][i] = 1.0;
      }
      // Loop backward through the derivative statements
      for (Offset ist = n_statements_-1; ist > 0; ist--) {
	const Statement& statement = statement_[ist];
	// We copy the RHS to "a" in case it appears on the LHS in any
	// of the following statements
	Real a[ADEPT_MULTIPASS_SIZE];
#if ADEPT_MULTIPASS_SIZE > 4
	// For large blocks, we only process the ones where a[i] is
	// non-zero
	Offset i_non_zero[ADEPT_MULTIPASS_SIZE];
#endif
	Offset n_non_zero = 0;
	for (Offset i = 0; i < ADEPT_MULTIPASS_SIZE; i++) {
	  a[i] = gradient_multipass_[statement.offset][i];
	  gradient_multipass_[statement.offset][i] = 0.0;
	  if (a[i] != 0.0) {
#if ADEPT_MULTIPASS_SIZE > 4
	    i_non_zero[n_non_zero++] = i;
#else
	    n_non_zero = 1;
#endif
	  }
	}
	// Only do anything for this statement if any of the a values
	// are non-zero
	if (n_non_zero) {
	  // Loop through the operations
	  for (Offset iop = statement_[ist-1].end_plus_one;
	       iop < statement.end_plus_one; iop++) {
	    // Try to minimize pointer dereferencing by making local
	    // copies
	    register Real multiplier = multiplier_[iop];
	    register Real* __restrict__ gradient_multipass 
	      = &(gradient_multipass_[offset_[iop]][0]);
#if ADEPT_MULTIPASS_SIZE > 4
	    // For large blocks, loop over only the indices
	    // corresponding to non-zero a
	    for (Offset i = 0; i < n_non_zero; i++) {
	      gradient_multipass[i_non_zero[i]] += multiplier*a[i_non_zero[i]];
	    }
#else
	    // For small blocks, do all indices
	    for (Offset i = 0; i < ADEPT_MULTIPASS_SIZE; i++) {
	      gradient_multipass[i] += multiplier*a[i];
	    }
#endif
	  }
	}
      } // End of loop over statement
      // Copy the gradients corresponding to the independent variables
      // into the Jacobian matrix
      for (Offset iindep = 0; iindep < n_independent(); iindep++) {
	for (Offset i = 0; i < ADEPT_MULTIPASS_SIZE; i++) {
	  jacobian_out[iindep*n_dependent()+i_dependent+i] 
	    = gradient_multipass_[independent_offset_[iindep]][i];
	}
      }
      i_dependent += ADEPT_MULTIPASS_SIZE;
    } // End of loop over blocks
    
    // Now do the same but for the remaining few rows in the matrix
    if (n_extra > 0) {
      zero_gradient_multipass();
      for (Offset i = 0; i < n_extra; i++) {
	gradient_multipass_[dependent_offset_[i_dependent+i]][i] = 1.0;
      }
      for (Offset ist = n_statements_-1; ist > 0; ist--) {
	const Statement& statement = statement_[ist];
	Real a[ADEPT_MULTIPASS_SIZE];
#if ADEPT_MULTIPASS_SIZE > 4
	Offset i_non_zero[ADEPT_MULTIPASS_SIZE];
#endif
	Offset n_non_zero = 0;
	for (Offset i = 0; i < n_extra; i++) {
	  a[i] = gradient_multipass_[statement.offset][i];
	  gradient_multipass_[statement.offset][i] = 0.0;
	  if (a[i] != 0.0) {
#if ADEPT_MULTIPASS_SIZE > 4
	    i_non_zero[n_non_zero++] = i;
#else
	    n_non_zero = 1;
#endif
	  }
	}
	if (n_non_zero) {
	  for (Offset iop = statement_[ist-1].end_plus_one;
	       iop < statement.end_plus_one; iop++) {
	    register Real multiplier = multiplier_[iop];
	    register Real* __restrict__ gradient_multipass 
	      = &(gradient_multipass_[offset_[iop]][0]);
#if ADEPT_MULTIPASS_SIZE > 4
	    for (Offset i = 0; i < n_non_zero; i++) {
	      gradient_multipass[i_non_zero[i]] += multiplier*a[i_non_zero[i]];
	    }
#else
	    for (Offset i = 0; i < n_extra; i++) {
	      gradient_multipass[i] += multiplier*a[i];
	    }
#endif
	  }
	}
      }
      for (Offset iindep = 0; iindep < n_independent(); iindep++) {
	for (Offset i = 0; i < n_extra; i++) {
	  jacobian_out[iindep*n_dependent()+i_dependent+i] 
	    = gradient_multipass_[independent_offset_[iindep]][i];
	}
      }
    }
    
    return SUCCESS;
  }
  
  
  // Compute the Jacobian matrix; note that jacobian_out must be
  // allocated to be of size m*n, where m is the number of dependent
  // variables and n is the number of independents. In the resulting
  // matrix, the "m" dimension of the matrix varies fastest. This is
  // implemented by calling one of jacobian_forward and
  // jacobian_reverse, whichever would be faster.
  Status
  Stack::jacobian(Real* jacobian_out)
  {
    if (n_independent() < n_dependent()) {
      return jacobian_forward(jacobian_out);
    }
    else {
      return jacobian_reverse(jacobian_out);
    }
  }
  
  // Print each derivative statement to standard output
  void
  Stack::print() const
  {
    for (Offset ist = 1; ist < n_statements_; ist++) {
      const Statement& statement = statement_[ist];
      std::cout << ist
		<< ": d[" << statement.offset
		<< "] = ";
      
      if (statement_[ist-1].end_plus_one == statement_[ist].end_plus_one) {
	std::cout << "0\n";
      }
      else {    
	for (Offset i = statement_[ist-1].end_plus_one;
	     i < statement.end_plus_one; i++) {
	  std::cout << " + " << multiplier_[i] << "*d[" << offset_[i] << "]";
	}
	std::cout << "\n";
      }
    }
  }
  
  // Print the current gradient list to standard output
  Status
  Stack::print_gradients() {
    if (gradients_are_initialized()) {
      for (Offset i = 0; i < max_gradient_; i++) {
	if (i%10 == 0) {
	  if (i != 0) {
	    std::cout << "\n";
	  }
	  std::cout << i << ":";
	}
	std::cout << " " << gradient_[i];
      }
      std::cout << "\n";
      return SUCCESS;
    }
    else {
      return FAILURE_GRADIENT_NOT_INITIALIZED;
    }
  }


#ifdef ADEPT_STACK_STORAGE_DYNAMIC
  // Initialize the vector of gradients ready for the adjoint
  // calculation
  void
  Stack::initialize_gradients() {
    if (max_gradient_ > 0) {
      if (n_allocated_gradients_ < max_gradient_) {
	if (gradient_) {
	  delete[] gradient_;
	}
	gradient_ = new Real[max_gradient_];
	n_allocated_gradients_ = max_gradient_;
      }
      for (Offset i = 0; i < max_gradient_; i++) {
	gradient_[i] = 0.0;
      }
    }
    gradients_initialized_ = true;
  }
#else
  void initialize_gradients() {
#ifdef ADEPT_STACK_STORAGE_STL
    gradient_.resize(max_gradient_+10, 0.0);
#else
    for (Offset i = 0; i < max_gradient_; i++) {
      gradient_[i] = 0.0;
    }
#endif
    gradient_initialized_ = true;
  }
#endif

  // Sending a Stack object to a stream reports information about the
  // stack
  std::ostream&
  operator<<(std::ostream& os, const Stack& stack)
  {
    os << "Automatic Differentiation Stack (";
    if ((!stack.is_thread_unsafe()) && _stack_current_thread == &stack) {
      os << "currently attached - thread safe):\n";
    }
    else if (stack.is_thread_unsafe() && _stack_current_thread_unsafe == &stack) {
      os << "currently attached - thread unsafe):\n";
    }
    else {
      os << "currently detached):\n";
    }
    os << "   " << stack.n_statements() << " statements (" 
       << stack.n_allocated_statements() << " allocated)\n";
    os << "   " << stack.n_operations() << " operations (" 
       << stack.n_allocated_operations() << " allocated)\n";
    os << "   " << stack.max_gradients() << " gradients needed (current index "
       << stack.i_gradient() << ")\n";
    if (stack.gradients_are_initialized()) {
      os << "   " << stack.max_gradients() << " gradients assigned (" 
	 << stack.n_allocated_gradients() << " allocated)\n";
    }
    else {
      os << "   0 gradients assigned (" << stack.n_allocated_gradients()
	 << " allocated)\n";
    }
    if (stack.gradients_allocated_suboptimally()) {
      os << "   Warning: gradients have been allocated sub-optimally because adouble objects\n"
	 << "   have not been destructed in reverse order to the order they were constructed.";
    }

    return os;
  }

} // End namespace adept

