/* adept.h -- Header for fast automatic differentiation using expression templates

    Copyright (C) 2012 Robin Hogan <r.j.hogan@reading.ac.uk>

    This file is part of the Adept library.

    This library is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ADEPT_H
#define ADEPT_H 1

/* Contents:
   SECTION 1: User changable defines
   SECTION 2: Defines requiring a library recompile 
   SECTION 3: Header files
   SECTION 4: Miscellaneous
   SECTION 5: Definition of Stack class
   SECTION 6: Definition of Expression types
   SECTION 7: Definition of Traits
   SECTION 8: Definition of aReal
   SECTION 9: Helper functions
*/


// ---------------------------------------------------------------------
// SECTION 1: User changable defines
// ---------------------------------------------------------------------

// The following can either be changed here, or define them just
// before including this header file in your code, or define using the
// -Dxxx compiler option.  These options to not need the library to be
// recompiled.

// A globally accessible stack needs to be present for arithmetic
// statements to access; by default this is thread safe but if you
// know you are running a single-threaded application then slightly
// faster performance may be achieved by defining this.
//#define ADEPT_STACK_THREAD_UNSAFE 1

// Define this to check whether the "multiplier" is zero before it is
// placed on the operation stack. This makes the forward pass slower
// and the reverse pass slightly faster, and is only worthwhile if
// many reverse passes will be carried out per forward pass (or if you
// have good reason to believe many variables in your code are zero).
// #define ADEPT_REMOVE_NULL_STATEMENTS 1

// If copy constructors for aReal objects are only used in the return
// values for functions then defining this will lead to slightly
// faster code, because it will be assumed that when a copy
// constructor is called the gradient_offset can simply be copied
// because the object being copied will shortly be destructed. You
// need to be sure that the code does not contain these constructions:
//   aReal x = y;
//   aReal x(y);
// where y is an aReal object.
//#define ADEPT_COPY_CONSTRUCTOR_ONLY_ON_RETURN_FROM_FUNCTION 1

// The initial size of the stacks, which can be grown if required
#ifndef ADEPT_INITIAL_STACK_LENGTH
#define ADEPT_INITIAL_STACK_LENGTH 1000
#endif


// ---------------------------------------------------------------------
// SECTION 2: Defines requiring a library recompile
// ---------------------------------------------------------------------

// The "stack" containing derivative information can be implemented in
// two ways: if ADEPT_STACK_STORAGE_STL is defined then C++ STL
// containers are used, otherwise dynamically allocated arrays are
// used.
//#define ADEPT_STACK_STORAGE_STL 1

// The number of rows/columns of a Jacobian that are calculated at
// once. The optimum value depends on platform. On modern Intel
// processors empirically the optimum appears to be 2 due to the SSE2
// instructions allowing two double multiplications to be carried out
// at once; presumably for single precision the optimum would be 4 due
// to the SSE instructions allowing four single-precision
// multiplications.
#ifndef ADEPT_MULTIPASS_SIZE
//#define ADEPT_MULTIPASS_SIZE 1
#define ADEPT_MULTIPASS_SIZE 2
//#define ADEPT_MULTIPASS_SIZE 4
//#define ADEPT_MULTIPASS_SIZE 8
//#define ADEPT_MULTIPASS_SIZE 16
//#define ADEPT_MULTIPASS_SIZE 32
#endif

namespace adept {
  // So far everything is double precision; a future version of this
  // library may permit mixed precision
  typedef double Real;

  // Used for all integer measures of array sizes and positions in an
  // array
  typedef unsigned int Offset;
}



// ---------------------------------------------------------------------
// SECTION 3: Header files
// ---------------------------------------------------------------------

#include <cmath>
#include <stdlib.h>
#include <iostream>
#include <typeinfo>
#include <utility>
#include <string>
#include <vector>

#ifdef ADEPT_STACK_STORAGE_STL
#include <valarray>
#endif


// ---------------------------------------------------------------------
// SECTION 4: Miscellaneous
// ---------------------------------------------------------------------

// If we are in the default case then provide
// ADEPT_STACK_STORAGE_DYNAMIC for future reference by the
// preprocessor
#ifndef ADEPT_STACK_STORAGE_STL
#define ADEPT_STACK_STORAGE_DYNAMIC 1
#endif

// This is sometimes used as a switch to turn on automatic
// differentiation; if we have included this file then it must be on
#ifndef AUTOMATIC_DIFFERENTIATION
#define AUTOMATIC_DIFFERENTIATION 1
#endif


namespace adept {
 
  // Declare a thread-safe and a thread-unsafe global pointer to the
  // current stack
  class Stack;
  extern __thread Stack* _stack_current_thread;
  extern Stack* _stack_current_thread_unsafe;

  // Define ADEPT_STACK to be the thread safe/unsafe version
#ifdef ADEPT_STACK_THREAD_UNSAFE
#define ADEPT_STACK _stack_current_thread_unsafe
#else
#define ADEPT_STACK _stack_current_thread
#endif

  // Structure describing the LHS of a derivative expression.  For dx
  // = z dy + y dz, "offset" would be the location of dx in the
  // gradient list, and "end_plus_one" would be one plus the location
  // of the final operation (multiplier-derivative pair) on the RHS,
  // in this case y dz.
  struct Statement {
    Statement() { }
    Statement(Offset offset_, Offset end_plus_one_)
      : offset(offset_), end_plus_one(end_plus_one_) { }
    Offset offset;
    Offset end_plus_one;
  };

  // Structure holding a fixed-size array of objects (intended for
  // double or float)
  template<int Size, class Type>
  struct Block {
    Block() { zero(); }
    const Type& operator[](Offset i) const { return data[i]; }
    Type& operator[](Offset i) { return data[i]; }
    void zero() { for (Offset i = 0; i < Size; i++) data[i] = 0.0; }
    Type data[Size];
  };

  // Return values from some functions
  typedef enum {
    SUCCESS = 0,
    FAILURE_GRADIENT_OUT_OF_RANGE,
    FAILURE_GRADIENT_NOT_INITIALIZED,
    FAILURE_NO_INITIAL_GRADIENTS,
    FAILURE_STACK_ALREADY_ACTIVE,
    FAILURE_XXDEPENDENTS_NOT_IDENTIFIED
  } Status;

  // Get a string representation of the error code
  std::string error_message(Status status);


  // ---------------------------------------------------------------------
  // SECTION 5: Definition of Stack class
  // ---------------------------------------------------------------------

  // Class containing derivative information of an algorithm, from which
  // the Jacobian matrix can be constructed, as well as tangent-linear
  // and adjoint operations being carried out for suitable input
  // derivatives.  Member functions not defined here are in
  // stack.cpp

  class Stack {
  public:
    // Only one constructor, which takes no arguments
    Stack() :
#ifdef ADEPT_STACK_STORAGE_DYNAMIC
      statement_(0), gradient_(0),
      multiplier_(0), offset_(0),
#endif
      n_statements_(0), n_allocated_statements_(0),
      n_operations_(0), n_allocated_operations_(0),
      i_gradient_(0), n_allocated_gradients_(0), max_gradient_(0),
      gradients_initialized_(false), 
      gradients_allocated_suboptimally_(false),
#ifdef ADEPT_STACK_THREAD_UNSAFE
      is_thread_unsafe_(true)
#else
      is_thread_unsafe_(false)
#endif
    { }
  
    // Destructor
    ~Stack();

    // Start recording the derivative information for an algorithm;
    // reset internal structures appropriately and preallocate "n"
    // statements and operations.
    Status start(Offset n = ADEPT_INITIAL_STACK_LENGTH);

    // Push an operation (i.e. a multiplier-gradient pair) on to the
    // stack.  We assume here that check_space() as been called before
    // so there is enough space to hold these elements.
  void push_rhs(const Real& multiplier, const Offset& gradient_offset) {
#ifdef ADEPT_REMOVE_NULL_STATEMENTS
    // If multiplier==0 then the resulting statement would have no
    // effect so we can speed up the subsequent adjoint/jacobian
    // calculations (at the expense of making this critical part of
    // the code slower)
    if (multiplier != 0.0) {
#endif

#ifdef ADEPT_STACK_STORAGE_STL
      multiplier_.push_back(multiplier);
      offset_.push_back(gradient_offset);
      n_operations_++;
#else
      multiplier_[n_operations_] = multiplier;
      offset_[n_operations_++] = gradient_offset;
#endif
      
#ifdef ADEPT_REMOVE_NULL_STATEMENTS
    }
#endif
  }

    // Push a statement on to the stack: this is done after a sequence
    // of operation pushes; gradient_offset is the offset of the
    // gradient on the LHS of the expression, while the "end_plus_one"
    // element is simply the current length of the operation list
    void push_lhs(const Offset& gradient_offset) {
#ifdef ADEPT_STACK_STORAGE_DYNAMIC
      if (n_statements_ >= n_allocated_statements_) {
	grow_statement_stack();
      }
#endif

#if ADEPT_STACK_STORAGE_STL
      statement_.push_back(Statement(gradient_offset, n_operations_));
      n_statements_++;
#else
      statement_[n_statements_].offset = gradient_offset;
      statement_[n_statements_++].end_plus_one = n_operations_;
#endif
    }

    // When an aReal object is created it is registered on the stack
    // and keeps a copy of its location, which is returned from this
    // function
    Offset register_gradient() {
      i_gradient_++;
      if (i_gradient_ > max_gradient_) {
	max_gradient_ = i_gradient_;
      }
      return i_gradient_-1;
    }

    // When an aReal object is destroyed it is unregistered from the
    // stack. If it is at the top of the stack then the stack pointer
    // can be decremented so that the space can be used by another
    // object. A gap can appear in the stack if an active object (or
    // array of active objects) is returned from a function, so we
    // need to keep track of a "gap" appearing in the stack. If the
    // user uses new and delete without any regard for this "last-in
    // first-out" preference then the number of gradients that are
    // allocated in the reverse pass may be larger than needed.
    void unregister_gradient(const Offset& gradient_offset) {
      if (gradient_offset+1 == i_gradient_) {
	i_gradient_--;
	if (i_gradient_ == igapend_gradient_) {
	  // We have unregistered the elements between the "gap" of
	  // unregistered element and the top of the stack, so can set
	  // the variables indicating the presence of the gap to zero
	  i_gradient_ = igapstart_gradient_;
	  igapstart_gradient_ = igapend_gradient_ = 0;
	}
      }
      else if (gradient_offset+1 == igapstart_gradient_) {
	// Removing a gradient from an existing gap
	igapstart_gradient_--;
      }
      else if (gradient_offset == igapend_gradient_) {
	// Not sure if this will ever happen: gap growing from top
	// upwards
	igapend_gradient_++;
      }
      else {
	// A new gap introduced
	if (igapstart_gradient_ != igapend_gradient_) {
	  // A gap has been "abandoned"
	  gradients_allocated_suboptimally_ = true;
	}
	igapstart_gradient_ = gradient_offset;
	igapend_gradient_ = gradient_offset+1;
      }
    }
  
    // Print each derivative statement to standard output
    void print() const;

    // Set the gradients in the list with offsets between start and
    // end_plus_one-1 to the values pointed to by "gradient"
    Status set_gradients(Offset start, Offset end_plus_one,
			 const Real* gradient) {
      // Need to initialize the gradient list if not already done
      if (!gradients_are_initialized()) {
	initialize_gradients();
      }
      if (end_plus_one > max_gradient_) {
	return FAILURE_GRADIENT_OUT_OF_RANGE;
      }
      for (Offset i = start, j = 0; i < end_plus_one; i++, j++) {
	gradient_[i] = gradient[j];
      }
      return SUCCESS;
    }

    // Get the gradients in the list with offsets between start and
    // end_plus_one-1 and put them in the location pointed to by
    // "gradient"
    Status get_gradients(Offset start, Offset end_plus_one,
			 Real* gradient) const {
      if (!gradients_are_initialized()) {
	return FAILURE_GRADIENT_NOT_INITIALIZED;
      }
      if (end_plus_one > max_gradient_) {
	return FAILURE_GRADIENT_OUT_OF_RANGE;
      }
      for (Offset i = start, j = 0; i < end_plus_one; i++, j++) {
	gradient[j] = gradient_[i];
      }
      return SUCCESS;
    }

    // Run the tangent-linear algorithm on the gradient list; normally
    // this call is preceded calls to set_gradient to load input
    // gradients and followed by calls to get_gradient to extract
    // gradients
    Status compute_tangent_linear();
    Status forward() { return compute_tangent_linear(); }

    // Run the adjoint algorithm on the gradient list; normally this
    // call is preceded calls to set_gradient to load input gradient
    // and followed by calls to get_gradient to extract gradient
    Status compute_adjoint();
    Status reverse() { return compute_adjoint(); }

    // Return the number of independent and dependent variables that
    // have been identified
    Offset n_independent() { return independent_offset_.size(); }
    Offset n_dependent() { return dependent_offset_.size(); }

    // Compute the Jacobian matrix; note that jacobian_out must be
    // allocated to be of size m*n, where m is the number of dependent
    // variables and n is the number of independents. The independents
    // and dependents must have already been identified with the
    // functions "independent" and "dependent", otherwise this
    // function will fail with FAILURE_XXDEPENDENT_NOT_IDENTIFIED. In
    // the resulting matrix, the "m" dimension of the matrix varies
    // fastest. This is implemented by calling one of jacobian_forward
    // and jacobian_reverse, whichever would be faster.
    Status jacobian(Real* jacobian_out);

    // Compute the Jacobian matrix, but explicitly specify whether
    // this is done with repeated forward or reverse passes.
    Status jacobian_forward(Real* jacobian_out);
    Status jacobian_reverse(Real* jacobian_out);

    // In order to compute the jacobian we need to first declare which
    // active variables are independent (x) and which are dependent
    // (y). First, the following two functions declare an individual
    // active variable and an array of active variables to be
    // independent. Note that we use templates here because aReal has
    // not been defined.
    template <class A>
    void independent(const A& x) {
      independent_offset_.push_back(x.gradient_offset());
    }
    template <class A>
    void independent(const A* x, Offset n) {
      for (Offset i = 0; i < n; i++) {
	independent_offset_.push_back(x[i].gradient_offset());
      }
    }

    // Likewise, delcare the dependent variables
    template <class A>
    void dependent(const A& x) {
      dependent_offset_.push_back(x.gradient_offset());
    }
    template <class A>
    void dependent(const A* x, Offset n) {
      for (Offset i = 0; i < n; i++) {
	dependent_offset_.push_back(x[i].gradient_offset());
      }
    }

    // Print the current gradient list to standard output
    Status print_gradients();

    // Clear the contents of the various lists (although leave the
    // memory allocated for a future call where possible)
    void clear() {
#ifdef ADEPT_STACK_STORAGE_STL
      statement_.clear();
      multiplier_.clear();
      offset_.clear();
#endif
      independent_offset_.clear();
      dependent_offset_.clear();
      gradient_multipass_.clear();
      n_operations_ = 0;
      n_statements_ = 0;
      i_gradient_ = 0;
      max_gradient_ = 1;
      gradients_initialized_ = false;
      gradients_allocated_suboptimally_ = false;
    }

    // This stack will stop being the one that aReal objects refer
    // to; this may be useful if the thread needs to use another stack
    // object for the next algorithm
    void detach() {
      if (ADEPT_STACK == this) {
	ADEPT_STACK = 0; 
      }
    }


    // Have the gradients been initialized?
    bool gradients_are_initialized() const { return gradients_initialized_; }

    // Return the number of statements, operations, and how much
    // memory has been allocated for each
    Offset n_statements() const { return n_statements_; }
    Offset n_allocated_statements() const { return n_allocated_statements_; }
    Offset n_operations() const { return n_operations_; }
    Offset n_allocated_operations() const { return n_allocated_operations_; }

    // Return the maximum number of gradients required to perform
    // adjoint calculation
    Offset max_gradients() const { return max_gradient_; }

    // Return the index to the current gradient
    Offset i_gradient() const { return i_gradient_; }

    // Return the number of gradients memory has been allocated for
    Offset n_allocated_gradients() const { return n_allocated_gradients_; }

    // Return the number of bytes used
    Offset memory() const {
      return n_statements()*sizeof(Offset)*2
	+ n_operations()*(sizeof(Real)+sizeof(Offset))
	+ max_gradients()*sizeof(Real);
    }

    bool is_thread_unsafe() const { return is_thread_unsafe_; }
    bool gradients_allocated_suboptimally() const { return gradients_allocated_suboptimally_; }

#ifdef ADEPT_STACK_STORAGE_DYNAMIC
    // Check whether the operation stack contains enough space for n
    // new operations; if not, grow it
    void check_space(const Offset& n) {
      if (n_allocated_operations_ < n_operations_+n+1) {
	grow_operation_stack(n);
      }
    }

private:
    // Grow the capacity of the operation or statement stacks to hold
    // a minimum of "min" elements. If min=0 then the stacks are
    // doubled in size.
    void grow_operation_stack(Offset min = 0);
    void grow_statement_stack(Offset min = 0);

    // Set to zero the gradients required by a Jacobian calculation
    void zero_gradient_multipass() {
      for (Offset i = 0; i < gradient_multipass_.size(); i++) {
	gradient_multipass_[i].zero();
      }
    }

#else // NOT ADEPT_STACK_STORAGE_DYNAMIC
    // For non-dynamic storage, check_space does nothing
    void check_space(const Offset& n) { }
#endif

  private:
    // Initialize the vector of gradients ready for the adjoint
    // calculation
    void initialize_gradients();

  private:
    // --- DATA SECTION ---

#ifdef ADEPT_STACK_STORAGE_STL
    // Data are stored using standard template library containers
    std::vector<Statement> statement_;
    std::valarray<Real> gradient_;
    std::vector<Real> multiplier_;
    std::vector<Offset> offset_;
#else
    // Data are stored as dynamically allocated arrays
    Statement* __restrict__ statement_ ;
    Real* __restrict__ gradient_;
    Real* __restrict__ multiplier_;
    Offset* __restrict__ offset_;
#endif
    // For Jacobians we process multiple rows/columns at once so need
    // what is essentially a 2D array
    std::vector<Block<ADEPT_MULTIPASS_SIZE,Real> > gradient_multipass_;
    // Offsets of the independent and dependent variables
    std::vector<Offset> independent_offset_;
    std::vector<Offset> dependent_offset_;

    Offset n_statements_;           // Number of statements
    Offset n_allocated_statements_; // Space allocated for statements
    Offset n_operations_;           // Number of operations
    Offset n_allocated_operations_; // Space allocated for statements
    Offset i_gradient_;             // Current number of gradients
    Offset n_allocated_gradients_;  // Number of allocated gradients
    Offset max_gradient_;           // Max number of gradients to store
    Offset igapstart_gradient_;     // Start of gap in allocated
				    // gradients (zero if no gap)
    Offset igapend_gradient_;       // End of gap in allocated gradients
    bool gradients_initialized_;    // Have the gradients been
				    // initialized?
    bool gradients_allocated_suboptimally_;
    const bool is_thread_unsafe_;
};


  // ---------------------------------------------------------------------
  // SECTION 6: Definition of Expression types
  // ---------------------------------------------------------------------

  // In this section the expression template approach is employed:
  // each operator and mathematical function returns a templated type,
  // rather than simply the numerical result of the operation

  // What type does the value() member function of each type of
  // expression return?
  //#define ADEPT_VALUE_RETURN_TYPE const Real&
#define ADEPT_VALUE_RETURN_TYPE Real

  class aReal;

  // The Expression type from which all other types of expression
  // derive. Each member function simply calls the specialized version
  // of the function according to the expression's true type, which is
  // given by its template argument.
  template<class A>
  struct Expression {
    // This indicates what type the expression resolves to, and allows
    // for future versions where there may be an underlying active
    // complex or single-precision type instead of double
    typedef aReal resolve_type;

    // Cast the expression to its true type, given by the template
    // argument
    const A& cast() const { return static_cast<const A&>(*this); }

    // Calculate the gradient of the mathematical operation that this
    // expression represents and pass the result to its argument (you
    // probably need to read the paper to understand why this is
    // useful). For functions f(a), pass df/da to the argument in the
    // first case and pass multiplier*df/da in the second case.
    void calc_gradient(Stack& stack) const {
      cast().calc_gradient(stack);
    }

    // As the previous but multiplying the gradient by "multiplier"
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      cast().calc_gradient(stack, multiplier);
    }

    // Return the numerical value of the expression
    ADEPT_VALUE_RETURN_TYPE value() const {
      return cast().value();
    }

    // Calculate the gradient and return the numerical value
    Real value_and_gradient(Stack& stack) const {
      cast().calc_gradient(stack);
      return cast().value();
    }

  private:
    // Intentionally inaccessible to prevent an expression appearing
    // on the left-hand-side of a statement
    Expression& operator=(const Expression&) { return *this; };
  };

  // Now define particular types of expression, using static
  // polymorphism via the Curiously Recurring Template Pattern

  // Add: an expression plus another expression
  template <class A, class B>
  struct Add : public Expression<Add<A,B> > {
    Add(const Expression<A>& a, const Expression<B>& b)
      : a_(a.cast()), b_(b.cast()) { }
    // If f(a,b) = a + b, df/da = 1 and likewise for df/db so simply
    // call a and b's versions of calc_gradient
    void calc_gradient(Stack& stack) const {
      a_.calc_gradient(stack);
      b_.calc_gradient(stack);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      a_.calc_gradient(stack, multiplier);
      b_.calc_gradient(stack, multiplier);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return a_.value() + b_.value();
    }
  private:
    // Store constant references to the arguments of the addition
    const A& a_;
    const B& b_;
  };

  // Overload the addition operator for Expression arguments to return
  // an Add type
  template <class A, class B>
  inline
  Add<A,B> operator+(const Expression<A>& a,
		     const Expression<B>& b) {
    return Add<A,B>(a.cast(),b.cast());
  }

  // Subtract: an expression minus another expression
  template <class A, class B>
  struct Subtract : public Expression<Subtract<A,B> > {
    Subtract(const Expression<A>& a, const Expression<B>& b)
      : a_(a.cast()), b_(b.cast()) { }
    // If f(a,b) = a - b, df/da = 1 and df/db = -1
    void calc_gradient(Stack& stack) const {
      a_.calc_gradient(stack);
      b_.calc_gradient(stack, -1.0);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      a_.calc_gradient(stack, multiplier);
      b_.calc_gradient(stack, -multiplier);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return a_.value() - b_.value();
    }
  private:
    const A& a_;
    const B& b_;
  };

  // Overload subtraction operator for Expression arguments
  template <class A, class B>
  inline
  Subtract<A,B> operator-(const Expression<A>& a,
			  const Expression<B>& b) {
    return Subtract<A,B>(a.cast(),b.cast());
  }

  // Multiply: an expression multiplied by another expression
  template <class A, class B>
  struct Multiply : public Expression<Multiply<A,B> > {
    Multiply(const Expression<A>& a, const Expression<B>& b)
      : a_(a.cast()), b_(b.cast()) { }
    // If f(a,b) = a*b, df/da = b and df/db = a
    void calc_gradient(Stack& stack) const {
      a_.calc_gradient(stack, b_.value());
      b_.calc_gradient(stack, a_.value());
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      a_.calc_gradient(stack, b_.value()*multiplier);
      b_.calc_gradient(stack, a_.value()*multiplier);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return a_.value() * b_.value();
    }
  private:
    const A& a_;
    const B& b_;
  };

  // Overload multiplication operator for Expression arguments
  template <class A, class B>
  inline
  Multiply<A,B> operator*(const Expression<A>& a,
			  const Expression<B>& b) {
    return Multiply<A,B>(a.cast(),b.cast());
  }

  // Divide: an expression divided by another expression
  template <class A, class B>
  struct Divide : public Expression<Divide<A,B> > {
    Divide(const Expression<A>& a, const Expression<B>& b)
      : a_(a.cast()), b_(b.cast()), one_over_b_(1.0/b_.value()),
	result_(a_.value()*one_over_b_) { }
    // If f(a,b) = a/b then df/da = 1/b and df/db = -a/(b*b)
    void calc_gradient(Stack& stack) const {
      a_.calc_gradient(stack, one_over_b_);
      b_.calc_gradient(stack, -result_*one_over_b_);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      Real tmp = multiplier*one_over_b_;
      a_.calc_gradient(stack, tmp);
      b_.calc_gradient(stack, -tmp*result_);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return result_;
    }
  private:
    const A& a_;
    const B& b_;
    Real one_over_b_;
    Real result_;
  };
  
  // Overload division operator for Expression arguments
  template <class A, class B>
  inline
  Divide<A,B> operator/(const Expression<A>& a,
			const Expression<B>& b) {
    return Divide<A,B>(a.cast(),b.cast());
  }

  // ScalarAdd: an expression plus a scalar
  template <class A>
  struct ScalarAdd : public Expression<ScalarAdd<A> > {
    ScalarAdd(const Expression<A>& a, const Real& b)
      : a_(a.cast()), result_(a_.value() + b) { }
    void calc_gradient(Stack& stack) const {
      a_.calc_gradient(stack);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      a_.calc_gradient(stack, multiplier);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return result_;
    }
  private:
    const A& a_;
    Real result_;
  };

  // Overload addition operator for expression plus scalar and scalar
  // plus expression
  template <class A>
  inline
  ScalarAdd<A> operator+(const Expression<A>& a,
			 const Real& b) {
    return ScalarAdd<A>(a.cast(),b);
  }
  template <class A>
  inline
  ScalarAdd<A> operator+(const Real& b,
			 const Expression<A>& a) {
    return ScalarAdd<A>(a.cast(),b);
  }

  // Overload subtraction operator for expression minus scalar to
  // return a ScalarAdd object
  template <class A>
  inline
  ScalarAdd<A> operator-(const Expression<A>& a,
			 const Real& b) {
    return ScalarAdd<A>(a.cast(),-b);
  }

  // ScalarSubtract: scalar minus expression
  template <class B>
  struct ScalarSubtract : public Expression<ScalarSubtract<B> > {
    ScalarSubtract(const Real& a, const Expression<B>& b)
      : b_(b.cast()), result_(a - b_.value()) {}
    void calc_gradient(Stack& stack) const {
      b_.calc_gradient(stack, -1.0);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      b_.calc_gradient(stack, -multiplier);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return result_;
    }
  private:
    const B& b_;
    Real result_;
  };

  // Overload subtraction operator for scalar minus expression
  template <class B>
  inline
  ScalarSubtract<B> operator-(const Real& a,
			      const Expression<B>& b) {
    return ScalarSubtract<B>(a, b.cast());
  }

  // ScalarMultiply: expression multiplied by scalar
  template <class A>
  struct ScalarMultiply
    : public Expression<ScalarMultiply<A> > {
    ScalarMultiply(const Expression<A>& a, const Real& b)
      : a_(a.cast()), b_(b) { }
    void calc_gradient(Stack& stack) const {
      a_.calc_gradient(stack, b_);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      a_.calc_gradient(stack, multiplier*b_);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return a_.value() * b_;
    }
  private:
    const A& a_;
    Real b_;
  };

  // Overload multiplication operator for expression multiplied by
  // scalar and scalar multiplied by expression
  template <class A>
  inline
  ScalarMultiply<A> operator*(const Expression<A>& a,
			      const Real& b) {
    return ScalarMultiply<A>(a.cast(),b);
  }
  template <class A>
  inline
  ScalarMultiply<A> operator*(const Real& b,
			      const Expression<A>& a) {
    return ScalarMultiply<A>(a.cast(),b);
  }
  
  // Overload division operator for expression divided by scalar to
  // return ScalarMultiply object
  template <class A>
  inline
  ScalarMultiply<A> operator/(const Expression<A>& a,
			      const Real& b) {
    return ScalarMultiply<A>(a.cast(),1.0/b);
  }

  // ScalarDivide: scalar divided by expression
  template <class B>
  struct ScalarDivide : public Expression<ScalarDivide<B> > {
    ScalarDivide(const Real& a, const Expression<B>& b)
      : b_(b.cast()), one_over_b_(1.0/b_.value()), 
	result_(a * one_over_b_) { }
    // If f(a,b) = a/b then df/db = -a/(b*b)    
    void calc_gradient(Stack& stack) const {
      b_.calc_gradient(stack, -result_*one_over_b_);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      b_.calc_gradient(stack, -multiplier*result_*one_over_b_);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return result_;
    }
  private:
    const B& b_;
    Real one_over_b_;
    Real result_;
  };

  // Overload division operator for scalar divided by expression
  template <class B>
  inline
  ScalarDivide<B> operator/(const Real& a,
			    const Expression<B>& b) {
    return ScalarDivide<B>(a,b.cast());
  }
  
  // Conditional operators should behave exactly the same as with
  // non-active arguments so in each of the cases below the value()
  // function is called to extract the value of the expression
#define ADEPT_DEFINE_CONDITIONAL(OPERATOR, OP)			\
  template <class A, class B>					\
  inline							\
  bool OPERATOR(const Expression<A>& a,				\
		const Expression<B>& b) {			\
    return a.value() OP b.value();				\
  }								\
								\
  template <class A>						\
  inline							\
  bool OPERATOR(const Expression<A>& a, const Real& b) {	\
    return a.value() OP b;					\
  }								\
  								\
  template <class B>						\
  inline							\
  bool OPERATOR(const Real& a, const Expression<B>& b) {	\
    return a OP b.value();					\
  }

  ADEPT_DEFINE_CONDITIONAL(operator==, ==)
  ADEPT_DEFINE_CONDITIONAL(operator!=, !=)
  ADEPT_DEFINE_CONDITIONAL(operator>, >)
  ADEPT_DEFINE_CONDITIONAL(operator<, <)
  ADEPT_DEFINE_CONDITIONAL(operator>=, >=)
  ADEPT_DEFINE_CONDITIONAL(operator<=, <=)
  
#undef ADEPT_DEFINE_CONDITIONAL
  
  // UnaryMinus: negation of expression 
  template <class A>
  struct UnaryMinus : public Expression<UnaryMinus<A> > {
    UnaryMinus(const Expression<A>& a)
      : a_(a.cast()) { }
    void calc_gradient(Stack& stack) const {
      a_.calc_gradient(stack, -1.0);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      a_.calc_gradient(stack, -multiplier);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return -a_.value();
    }
  private:
    const A& a_;
  };
  
  // Overload unary minus of expression
  template <class A>
  inline
  UnaryMinus<A> operator-(const Expression<A>& a) {
    return UnaryMinus<A>(a.cast());
  }

  // Unary plus: returns the argument
  template <class A>
  inline
  A operator+(const Expression<A>& a) {
    return a;
  }

  // Exponential of an expression
  template <class A>
  struct Exp : public Expression<Exp<A> > {
    Exp(const Expression<A>& a)
      : a_(a.cast()), result_(exp(a.value())) { }
    // If f(a) = exp(a) then df/da = exp(a)
    void calc_gradient(Stack& stack) const {
      a_.calc_gradient(stack, result_);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      a_.calc_gradient(stack, result_*multiplier);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return result_;
    }
  private:
    const A& a_;
    Real result_;
  };
  
} // End namespace adept

// It is important to place overloads of mathematical functions in the
// global namespace.  If exp(Expression) was placed in the adept
// namespace then the Exp class (which is in the adept namespace would
// not be able to find the std::exp(double) function due to C++ name
// look-up rules.

// Overload exp for Expression objects
template <class A>
inline
adept::Exp<A> exp(const adept::Expression<A>& a) {
  return adept::Exp<A>(a.cast());
}


// Enable unary mathematical functions when the derivative is most
// easily written in terms of the argument of the function; note that
// the class is in the adept name space but the function is not.
#define ADEPT_DEFINE_UNARY_FUNCTION(OP,FUNC,DERIVATIVE)		\
  namespace adept {						\
    template <class A>						\
    struct OP : public Expression<OP<A> > {			\
      OP(const Expression<A>& a)				\
	: a_(a.cast()), result_(FUNC(a_.value())) { }		\
      void calc_gradient(Stack& stack) const {			\
	a_.calc_gradient(stack, DERIVATIVE);			\
      }								\
      void calc_gradient(Stack& stack,				\
			 const Real& multiplier) const {	\
	a_.calc_gradient(stack, (DERIVATIVE)*multiplier);	\
      }								\
      ADEPT_VALUE_RETURN_TYPE value() const {			\
	return result_;						\
      }								\
    private:							\
    const A& a_;						\
    Real result_;						\
    };								\
  }								\
  template <class A>						\
  inline							\
  adept:: OP<A> FUNC(const adept::Expression<A>& a) {		\
    return adept:: OP<A>(a.cast());				\
  }

// The final argument is the value of the derivative
ADEPT_DEFINE_UNARY_FUNCTION(Log,  log,  1.0/a_.value())
ADEPT_DEFINE_UNARY_FUNCTION(Log10,log10,0.434294481903252/a_.value())
ADEPT_DEFINE_UNARY_FUNCTION(Sin,  sin,  cos(a_.value()))
ADEPT_DEFINE_UNARY_FUNCTION(Cos,  cos,  -sin(a_.value()))
ADEPT_DEFINE_UNARY_FUNCTION(Asin, asin, 1.0/sqrt(1.0-a_.value()*a_.value()))
ADEPT_DEFINE_UNARY_FUNCTION(Acos, acos, -1.0/sqrt(1.0-a_.value()*a_.value()))
ADEPT_DEFINE_UNARY_FUNCTION(Atan, atan, 1.0/(1+a_.value()*a_.value()))
ADEPT_DEFINE_UNARY_FUNCTION(Sinh, sinh, cosh(a_.value()))
ADEPT_DEFINE_UNARY_FUNCTION(Cosh, cosh, sinh(a_.value()))
ADEPT_DEFINE_UNARY_FUNCTION(Abs,  abs,  (a_.value()>0.0)-(a_.value()<0.0))

#undef ADEPT_DEFINE_UNARY_FUNCTION

// Need both fabs and abs for C compatibility, so get fabs to return
// an Abs object
template <class A>
inline
adept::Abs<A> fabs(const adept::Expression<A>& a) {
  return adept::Abs<A>(a.cast());
}

// Need to add ceil, floor...
// Lots more math function in math.h: erf, bessel functions etc...

// Enable mathematical functions when the derivative is most easily
// written in terms of the return value of the function
#define ADEPT_DEFINE_UNARY_FUNCTION2(OP,FUNC,DERIVATIVE)	\
  namespace adept {						\
    template <class A>						\
    struct OP : public Expression<OP<A> > {			\
      OP(const Expression<A>& a)				\
	: a_(a.cast()), result_(FUNC(a.value())) { }		\
      void calc_gradient(Stack& stack) const {			\
	a_.calc_gradient(stack, DERIVATIVE);			\
      }								\
      void calc_gradient(Stack& stack,				\
			 const Real& multiplier) const {	\
	a_.calc_gradient(stack, (DERIVATIVE)*multiplier);	\
      }								\
      ADEPT_VALUE_RETURN_TYPE value() const {			\
	return result_;						\
      }								\
    private:							\
    const A& a_;						\
    Real result_;						\
    };								\
  }								\
  template <class A>						\
  inline							\
  adept:: OP<A> FUNC(const adept::Expression<A>& a) {		\
    return adept:: OP<A>(a.cast());				\
  }

// Only sqrt in this category...
ADEPT_DEFINE_UNARY_FUNCTION2(Sqrt, sqrt, 0.5/result_)

#undef ADEPT_DEFINE_UNARY_FUNCTION2


// Enable mathematical functions when derivative is a little more
// complicated
#define ADEPT_DEFINE_UNARY_FUNCTION3(OP,FUNC,DERIV1, DERIV2)	\
  namespace adept {						\
    template <class A>						\
    struct OP : public Expression<OP<A> > {			\
      OP(const Expression<A>& a)				\
	: a_(a.cast()), result_(FUNC(a_.value())) { }		\
      void calc_gradient(Stack& stack) const {			\
	DERIV1;							\
	a_.calc_gradient(stack, DERIV2);			\
      }								\
      void calc_gradient(Stack& stack,				\
			 const Real& multiplier) const {	\
	DERIV1;							\
	a_.calc_gradient(stack, (DERIV2)*multiplier);		\
      }								\
      ADEPT_VALUE_RETURN_TYPE value() const {			\
	return result_;						\
      }								\
    private:							\
    const A& a_;						\
    Real result_;						\
    };								\
  }								\
  template <class A>						\
  inline							\
  adept:: OP<A> FUNC(const adept::Expression<A>& a) {		\
    return adept:: OP<A>(a.cast());				\
  }

ADEPT_DEFINE_UNARY_FUNCTION3(Tan,  tan, Real tmp = 1/cos(a_.value()), tmp * 2.0)
ADEPT_DEFINE_UNARY_FUNCTION3(Tanh, tanh, Real e=exp(2.0*a_.value()), 4.0*e/((2.0+2)*e+1.0))

#undef ADEPT_DEFINE_UNARY_FUNCTION3

namespace adept {
  // Pow: an expression to the power of another expression
  template <class A, class B>
  struct Pow : public Expression<Pow<A,B> > {
    Pow(const Expression<A>& a, const Expression<B>& b)
      : a_(a.cast()), b_(b.cast()), 
	result_(pow(a_.value(), b_.value())) { };
    // If f(a,b)=pow(a,b) then df/da=b*pow(a,b-1) and df/db=log(a)*pow(a,b)
    void calc_gradient(Stack& stack) const {
      a_.calc_gradient(stack, b_.value()*pow(a_.value(),b_.value()-1.0));
      b_.calc_gradient(stack, log(a_.value())*result_);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      a_.calc_gradient(stack, b_.value()*pow(a_.value(), 
					     b_.value()-1.0)*multiplier);
      b_.calc_gradient(stack, log(a_.value())*result_*multiplier);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return result_;
    }
  private:
    const A& a_;
    const B& b_;
    Real result_;
  };
  
  // PowScalarExponent: an expression to the power of a scalar
  template <class A>
  struct PowScalarExponent : public Expression<PowScalarExponent<A> > {
    PowScalarExponent(const Expression<A>& a, const Real& b)
      : a_(a.cast()), b_(b), result_(pow(a_.value(), b)) {}
    void calc_gradient(Stack& stack) const {
      a_.calc_gradient(stack, b_*pow(a_.value(),b_-1.0));
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      a_.calc_gradient(stack, b_*pow(a_.value(), b_-1.0)*multiplier);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return result_;
    }
  private:
    const A& a_;
    Real b_;
    Real result_;
  };

  // PowScalarBase: a scalar to the power of an expression
  template <class B>
  struct PowScalarBase : public Expression<PowScalarBase<B> > {
    PowScalarBase(const Real& a, const Expression<B>& b)
      : a_(a), b_(b.cast()), result_(pow(a_, b_.value())) { }
    void calc_gradient(Stack& stack) const {
      b_.calc_gradient(stack, log(a_)*result_);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      b_.calc_gradient(stack, log(a_)*result_*multiplier);
    }
    ADEPT_VALUE_RETURN_TYPE value() const {
      return result_;
    }
  private:
    Real a_;
    const B& b_;
    Real result_;
  };
} // End namespace adept

// Overload pow for Expression arguments
template <class A, class B>
inline
adept::Pow<A,B> pow(const adept::Expression<A>& a,
		    const adept::Expression<B>& b) {
  return adept::Pow<A,B>(a.cast(),b.cast());
}

// Overload pow for expression to the power of scalar
template <class A>
inline
adept::PowScalarExponent<A> pow(const adept::Expression<A>& a, 
				const adept::Real& b) {
  return adept::PowScalarExponent<A>(a.cast(),b);
}

// Overload pow for scalar to the power of an expression
template <class B>
inline
adept::PowScalarBase<B> pow(const adept::Real& a,
			    const adept::Expression<B>& b) {
  return adept::PowScalarExponent<B>(a, b.cast());
}

// If any further functions are added, remember to define a matching
// Trait specialization


// ---------------------------------------------------------------------
// SECTION 7: Definition of Traits
// ---------------------------------------------------------------------

namespace adept {
  // We want Traits<A>::n_active_variables to be defined at compile
  // time as the number of active variables in a particular expression
  // type A. This is done by defining a Traits type and specializing
  // it for each of the classes representing types of expression.

  template<class A>
  struct Traits {
    // This will not compile if the relevant trait is not defined for
    // a particular function, because n_active_variables is
    // intentionally not present
  };
  
  // aReal is an active variable so n_active_variables is 1
  template<>
  struct Traits<aReal> {
    enum { n_active_variables = 1 }; 
  };
  
  // For an arbitrary binary operation, the number of active variables
  // is the sum of the number of active variables in the two arguments
#define ADEPT_DEFINE_BINARY_TRAIT(OP)			\
  template<class A, class B>				\
  struct Traits<OP<A,B> > {				\
    enum { n_active_variables				\
	   = Traits<A>::n_active_variables		\
	   + Traits<B>::n_active_variables };		\
  };
  ADEPT_DEFINE_BINARY_TRAIT(Add)
  ADEPT_DEFINE_BINARY_TRAIT(Subtract)
  ADEPT_DEFINE_BINARY_TRAIT(Multiply)
  ADEPT_DEFINE_BINARY_TRAIT(Divide)
  ADEPT_DEFINE_BINARY_TRAIT(Pow)
  
#undef ADEPT_DEFINE_BINARY_TRAIT

  // For an arbitrary unary operation, the number of active variables
  // is equal to that for the argument
#define ADEPT_DEFINE_UNARY_TRAIT(OP)			\
  template<class A>					\
  struct Traits<OP<A> > {				\
    enum { n_active_variables				\
	   = Traits<A>::n_active_variables };		\
  };

  ADEPT_DEFINE_UNARY_TRAIT(ScalarAdd)
  ADEPT_DEFINE_UNARY_TRAIT(ScalarSubtract)
  ADEPT_DEFINE_UNARY_TRAIT(ScalarMultiply)
  ADEPT_DEFINE_UNARY_TRAIT(ScalarDivide)
  ADEPT_DEFINE_UNARY_TRAIT(UnaryMinus)
  ADEPT_DEFINE_UNARY_TRAIT(Exp)
  ADEPT_DEFINE_UNARY_TRAIT(Log)
  ADEPT_DEFINE_UNARY_TRAIT(Log10)
  ADEPT_DEFINE_UNARY_TRAIT(Sqrt)
  ADEPT_DEFINE_UNARY_TRAIT(Sin)
  ADEPT_DEFINE_UNARY_TRAIT(Cos)
  ADEPT_DEFINE_UNARY_TRAIT(Asin)
  ADEPT_DEFINE_UNARY_TRAIT(Acos)
  ADEPT_DEFINE_UNARY_TRAIT(Atan)
  ADEPT_DEFINE_UNARY_TRAIT(Sinh)
  ADEPT_DEFINE_UNARY_TRAIT(Cosh)
  ADEPT_DEFINE_UNARY_TRAIT(Tanh)
  ADEPT_DEFINE_UNARY_TRAIT(Abs)
  ADEPT_DEFINE_UNARY_TRAIT(PowScalarExponent)
  ADEPT_DEFINE_UNARY_TRAIT(PowScalarBase)
  
#undef DEFINE_UNARY_TRAIT


  // ---------------------------------------------------------------------
  // SECTION 8: Definition of aReal
  // ---------------------------------------------------------------------
  
  // The basic active type wrapping a Real (invariably double) number;
  // sometime this will be generalized to floats, complex<double>
  // etc. This inherits from Expression so that it can be used in
  // expressions.
  class aReal : public Expression<aReal> {
  public:
    // Constructor registers the new aReal object with the currently
    // active stack.  Note that this object is not explicitly
    // initialized with a particular number; the user should not
    // assume that it is set to zero but should later assign it to a
    // particular value. Otherwise in the reverse pass the
    // corresponding gradient will not be set to zero.
    aReal()
      : val_(0.0), gradient_offset_(ADEPT_STACK->register_gradient()) { }
    
    // This constructor is invoked with either of the following:
    //   aReal x = 1.0;
    //   aReal x(1.0);
    aReal(const Real& rhs)
      : val_(rhs), gradient_offset_(ADEPT_STACK->register_gradient())
    {
      // By pushing this to the statement stack without pushing
      // anything on to the operation stack we ensure that in the
      // reverse pass the gradient of this object will be set to zero
      // after it has been manipulated. This is important because the
      // gradient entry might be reused.
      ADEPT_STACK->push_lhs(gradient_offset_);
    }
    

#ifndef ADEPT_COPY_CONSTRUCTOR_ONLY_ON_RETURN_FROM_FUNCTION
    // Normal copy construction: register the new object then treat
    // this as an assignment
    aReal(const aReal& rhs) 
      : val_(0.0), gradient_offset_(ADEPT_STACK->register_gradient())
    {
      *this = rhs;
    }
#else
    // If copy constructors for aReal objects are only used in the
    // return values for functions then we may make a slight
    // optimization by simply copying the gradient_offset, since we
    // can expect that the object being copied will shortly be
    // destructed. This will only lead to the right answer if your
    // code does not contain either of the following constructions:
    //   aReal x = y;
    //   aReal x(y);
    // where y is an aReal object.
    aReal(const aReal& rhs)
      : val_(rhs.value()), gradient_offset_(rhs.gradient_offset()) { }
#endif
    
    // Construction with an expression
    template<class R>
    aReal(const Expression<R>& rhs)
      : gradient_offset_(ADEPT_STACK->register_gradient())
    {
      // Check there is enough space in the operation stack
      ADEPT_STACK->check_space(Traits<R>::n_active_variables);
      // Get the value and push the gradients on to the operation
      // stack, thereby storing the right-hand-side of the statement
      val_ = rhs.value_and_gradient(*ADEPT_STACK);
      // Push the gradient offet of this object on to the statement
      // stack, thereby storing the left-hand-side of the statement
      ADEPT_STACK->push_lhs(gradient_offset_);
    }
    
    // Destructor simply unregisters the object from the stack,
    // freeing up the gradient offset for another
    ~aReal() {
      ADEPT_STACK->unregister_gradient(gradient_offset_);
    }

    // Assignment operator with an active variable on the rhs
    aReal& operator=(const aReal& rhs) {
      // Check there is space in the operation stack for one more
      // entry
      ADEPT_STACK->check_space(1);
      // Same as construction with an expression (defined above)
      val_ = rhs.value_and_gradient(*ADEPT_STACK);
      ADEPT_STACK->push_lhs(gradient_offset_);
      return *this;
    }
    
    // Assignment operator with an inactive variable on the rhs
    aReal& operator=(const Real& rhs) {
      val_ = rhs;
      // Pushing the gradient offset on to the statement stack with no
      // corresponding operations ensures that the gradient will be
      // set to zero in the reverse pass when it is finished with
      ADEPT_STACK->push_lhs(gradient_offset_);
      return *this;
    }

    // Assignment operator with an expression on the rhs: very similar
    // to construction with an expression (defined above)
    template<class R>
    aReal& operator=(const Expression<R>& rhs) {
      ADEPT_STACK->check_space(Traits<R>::n_active_variables);
      val_ = rhs.value_and_gradient(*ADEPT_STACK);
      ADEPT_STACK->push_lhs(gradient_offset_);
      return *this;
    }

    // All the compound assignment operators are unpacked, i.e. a+=b
    // becomes a=a+b; first for an Expression on the rhs
    template<class R>
    aReal& operator+=(const Expression<R>& rhs) {
      return *this = (*this + rhs);
    }
    template<class R>
    aReal& operator-=(const Expression<R>& rhs) {
      return *this = (*this - rhs);
    }
    template<class R>
    aReal& operator*=(const Expression<R>& rhs) {
      return *this = (*this * rhs);
    }
    template<class R>
    aReal& operator/=(const Expression<R>& rhs) {
      return *this = (*this / rhs);
    }

    // And likewise for a scalar on the rhs
    aReal& operator+=(const Real& rhs) {
      val_ += rhs;
      return *this;
    }
    aReal& operator-=(const Real& rhs) {
      val_ -= rhs;
      return *this;
    }
    aReal& operator*=(const Real& rhs) {
      return *this = (*this * rhs);
    }
    aReal& operator/=(const Real& rhs) {
      return *this = (*this / rhs);
    }
    
    // If an expression leads to calc_gradient being called on an
    // aReal object, we push the multiplier and the gradient offset on
    // to the operation stack (or 1.0 if no multiplier is specified
    void calc_gradient(Stack& stack) const {
      stack.push_rhs(1.0, gradient_offset_);
    }
    void calc_gradient(Stack& stack, const Real& multiplier) const {
      stack.push_rhs(multiplier, gradient_offset_);
    }
   
    // Get the actual value of this object; needed for fprintf calls,
    // for example, to get the underlying Real type out
    ADEPT_VALUE_RETURN_TYPE value() const {
      return val_; 
    }

    // Push the gradient on to the operation stack and return the
    // numerical value
    Real value_and_gradient(Stack& stack) const { 
      stack.push_rhs(1.0, gradient_offset_);
      return val_;
    };

    // Get the offset of the gradient information for this object
    const Offset& gradient_offset() const { return gradient_offset_; }
 
    // Set the value 
    void set_value(const Real& x) { val_ = x; }

    // Set the value of the gradient, for initializing an adjoint;
    // note that the value of the gradient is not held in the aReal
    // object but rather held by the stack
    Status set_gradient(const Real& gradient) const {
      return ADEPT_STACK->set_gradients(gradient_offset_,
					gradient_offset_+1, &gradient);
    }

    // Get the value of the gradient, for extracting the adjoing after
    // calling reverse() on the stack
    Status get_gradient(Real& gradient) const {
      return ADEPT_STACK->get_gradients(gradient_offset_,
					gradient_offset_+1, &gradient);
    }
    Real get_gradient() const {
      Real gradient;
      ADEPT_STACK->get_gradients(gradient_offset_,
				 gradient_offset_+1, &gradient);
      return gradient;
    }
    
  private:
    // --- DATA SECTION ---
    Real val_;                     // The numerical value
    const Offset gradient_offset_; // Index to where the corresponding
				   // gradient will be held during the
				   // adjoint calculation
  }; // End of definition of aReal

  // This is what we called it in the paper
  typedef aReal adouble;


  // ---------------------------------------------------------------------
  // SECTION 9: Helper functions
  // ---------------------------------------------------------------------


  // When we need source files that will be compiled twice, both
  // including automatic differentiation and without (the latter case
  // with aReal typedef'ed to double), it is useful to be able to
  // access the underlying value for calls to fprintf etc, so we
  // define the value function which does this for expressions, and in
  // the case without automatic differentiation (in which this header
  // file would not be included), the value function would be defined
  // for double arguments to simply return the argument.
  template<class A>
  inline Real value(const Expression<A>& x) { return x.value(); }
  
  // A way of setting the initial values of an array of n aReal
  // objects without the expense of placing them on the stack
  template<typename Type>
  inline
  void set_values(aReal* a, Offset n, const Type* data)
  {
    for (Offset i = 0; i < n; i++) {
      a[i].set_value(data[i]);
    }
  }

  // Extract the values of an array of n aReal objects
  template<typename Type>
  inline
  void get_values(const aReal* a, Offset n, Type* data)
  {
    for (Offset i = 0; i < n; i++) {
      data[i] = a[i].value();
    }
  }
  
  // Set the initial gradients of an array of n aReal objects; this
  // should be done after the algorithm has called and before the
  // Stack::forward or Stack::reverse functions are called
  template<typename Type>
  inline
  void set_gradients(aReal* a, Offset n, const Type* data)
  {
    for (Offset i = 0; i < n; i++) {
      a[i].set_gradient(data[i]);
    }
  }
  
  // Extract the gradients from an array of aReal objects after the
  // Stack::forward or Stack::reverse functions have been called
  template<typename Type>
  inline
  void get_gradients(const aReal* a, Offset n, Type* data)
  {
    for (Offset i = 0; i < n; i++) {
      a[i].get_gradient(data[i]);
    }
  }
  
  // If we send an aReal object to a stream it behaves just like a
  // double
  inline
  std::ostream&
  operator<<(std::ostream& os, const adept::aReal& x)
  {
    os << value(x);
    return os;
  }
  
  // Sending a Stack object to a stream reports information about the
  // stack
  std::ostream& operator<<(std::ostream& os, const adept::Stack& stack);
  
} // End of namespace adept



#endif
